package ca.mcgill.sable.abc.impact.visualizer.model;

import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.Map.Entry;

import org.eclipse.core.resources.IFile;
import org.eclipse.core.resources.IProject;
import org.eclipse.core.resources.IResource;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.jdt.core.ICompilationUnit;
import org.eclipse.jdt.core.IJavaElement;
import org.eclipse.jdt.core.IJavaProject;
import org.eclipse.jdt.core.IMember;
import org.eclipse.jdt.core.ISourceRange;
import org.eclipse.jdt.core.JavaCore;
import org.eclipse.jface.text.ITextSelection;
import org.eclipse.ui.texteditor.MarkerUtilities;

import polyglot.util.Position;
import soot.RefType;
import soot.SootClass;
import soot.Type;
import abc.impact.impact.AdviceImpact;
import abc.impact.impact.AspectImpact;
import abc.impact.impact.InAspectImpact;
import abc.impact.impact.LookupChange;
import abc.impact.impact.LookupImpact;
import abc.impact.impact.Mutation;
import abc.impact.impact.ShadowingImpact;
import abc.impact.impact.StateImpact;
import abc.weaving.aspectinfo.AdviceDecl;
import abc.weaving.aspectinfo.Aspect;
import abc.weaving.aspectinfo.InAspect;
import ca.mcgill.sable.abc.impact.visualizer.actions.RunImpactAnalysisActionDelegate;
import ca.mcgill.sable.abc.impact.visualizer.model.IActPosition.RevealData;
import ca.mcgill.sable.abc.impact.visualizer.util.ImpactConstants;

public final class ImpactFactory {

	//prevent instantiation
	private ImpactFactory() {}

	private static ITreeNode invisibleRoot = new InvisibleRootTreeNode();
	private static List<ITreeNode> showNodes = new LinkedList<ITreeNode>();
	private static List<ITreeNode> selectedNodes = new LinkedList<ITreeNode>();
	
	public static Object getTreeData() {
		return invisibleRoot; 
	}
	
	public static List<ITreeNode> getShowData() {
		return showNodes;
	}
	
	private static class SelectVisitor extends AbstractTreeVisitor {
		
		private final ITextSelection selection;
		private final List<ITreeNode> selectedNodes;
		private final IFile file;
		
		public SelectVisitor(ITreeNode root, final IFile file, final ITextSelection selection) {
			super(root);
			this.file = file;
			this.selection = selection;
			selectedNodes = new LinkedList<ITreeNode>();
		}
		
		public List<ITreeNode> getSelectedNodes() {
			return selectedNodes;
		}

		@Override
		protected void action(ITreeNode node) {
			if (node instanceof IActJavaElement) {
				IActJavaElement jnode = (IActJavaElement) node;
				IJavaElement jEle = jnode.getJavaElement();
				if (jEle instanceof IMember) {
					IMember jMem = (IMember) jEle;
					try {
						ICompilationUnit cu = jMem.getCompilationUnit();
						IResource cuResource = cu.getCorrespondingResource();
						if(cu.getCorrespondingResource()!= null && cuResource.exists() 
								&& ((IFile)cuResource).equals(file)) {
							ISourceRange range = jMem.getNameRange();
							int selStart = selection.getOffset();
							int selEnd = selection.getLength() + selStart;
							int rangeStart = range.getOffset();
							int rangeEnd = range.getLength() + rangeStart;
							if (selStart >= rangeStart && selStart <= rangeEnd
									&& selEnd >= rangeStart && selEnd <= rangeEnd) {
								selectedNodes.add(node);
							}
						}
					} catch (CoreException e) {
						e.printStackTrace();
					}
				}
			}
			if (node instanceof IActPosition) {
				IActPosition pnode = (IActPosition) node;
				RevealData rData = pnode.getRevealData();
				IFile rFile = rData.getFile();
				if (rFile != null && rFile.exists() && rFile.equals(file)) {
					Position pos = rData.getPosition();
					int selStart = selection.getStartLine();
					int selEnd = selection.getEndLine();
					int posStart = pos.line() - 1;
					int posEnd = pos.endLine() - 1;
					if (selStart >= posStart && selStart <= posEnd
							&& selEnd >= posStart && selEnd <= posEnd) {
						selectedNodes.add(node);
					}
				}
			}
		}
	}
	
	/**
	 * Caller should call breadthTraverse()
	 * @author Dehua Zhang
	 *
	 */
	private static class DescendantsVisitor extends AbstractTreeVisitor {
		private final List<ITreeNode> descendantsNodes;
		private final List<ITreeNode> selectedNodes;
		
		public DescendantsVisitor(ITreeNode root, final List<ITreeNode> selectedNodes) {
			super(root);
			this.selectedNodes = selectedNodes;
			descendantsNodes =  new LinkedList<ITreeNode>();
		}
		
		public List<ITreeNode> getDescendantsNodes() {
			return descendantsNodes;
		}

		@Override
		protected void action(ITreeNode node) {
			if (descendantsNodes.contains(node.getParent()) ||
					selectedNodes.contains(node.getParent())) {
				descendantsNodes.add(node);
			}
		}

		@Override
		public void depthPostOrderTraverse() {
			throw new UnsupportedOperationException("This traverse method will not work.");
		}

		@Override
		public void depthPreOrderTraverse() {
			throw new UnsupportedOperationException("This traverse method will not work.");
		}
		
	}
	
	public static void createShowData(final IFile file, final ITextSelection selection) {
		
		showNodes = new LinkedList<ITreeNode>();
		
		SelectVisitor svisitor = new SelectVisitor(invisibleRoot, file, selection);
		svisitor.breadthTraverse();
		final List<ITreeNode> selected = svisitor.getSelectedNodes();
		selectedNodes = new LinkedList<ITreeNode>(selected);
		showNodes.addAll(selected);
		
		// add ancestors, allow duplication
		final List<ITreeNode> ancestors = new LinkedList<ITreeNode>();
		for (ITreeNode node : selected) {
			ITreeNode temp = node.getParent();
			while (temp != null) {
				ancestors.add(temp);
				temp = temp.getParent();
			}
		}
		showNodes.addAll(ancestors);
		
		// add descendants, 
		DescendantsVisitor dvisitor = new DescendantsVisitor(invisibleRoot, selected);
		dvisitor.breadthTraverse();
		final List<ITreeNode> descendants = dvisitor.getDescendantsNodes();
		showNodes.addAll(descendants);
		
	}
	
	/**
	 * Calculate the tree for view and markers
	 * Should be called after successful run of impact analysis
	 */
	public static void populateTreeAndMarker() {
		IProject project= RunImpactAnalysisActionDelegate.getCurrentProject();
		if (project != null) {
			createTreeData(project);
			createAllMarkers(project);
		}
	}
	
	private static void createAllMarkers(IProject project) {
		AbstractTreeVisitor visitor = new AbstractTreeVisitor(invisibleRoot) {
//			private List<ISourceRange> r = new ArrayList<ISourceRange>();
//			private List<Integer> l = new ArrayList<Integer>(); 
			
			@Override
			protected void action(ITreeNode node) {
				if (node instanceof IActJavaElement) {
					IActJavaElement jnode = (IActJavaElement) node;
					IJavaElement jEle = jnode.getJavaElement();
					if (jEle instanceof IMember) {
						IMember jMem = (IMember) jEle;
						try {
							ICompilationUnit cu = jMem.getCompilationUnit();
							if(cu.getResource()!= null && cu.getResource().exists()) {
								ISourceRange range = jMem.getNameRange();
//								if (! r.contains(range)) {
//									r.add(range);
									Map map = new HashMap();
									MarkerUtilities.setCharStart(map, range.getOffset());
									MarkerUtilities.setCharEnd(map, range.getOffset() + range.getLength());
									MarkerUtilities.setMessage(map, "impact");
									MarkerUtilities.createMarker(cu.getResource(), map, ImpactConstants.IMPACT_MARKER);
//								}
							}
						} catch (CoreException e) {
							e.printStackTrace();
						}
					}
				}
				if (node instanceof IActPosition) {
					IActPosition pnode = (IActPosition) node;
					RevealData rData = pnode.getRevealData();
					IFile file = rData.getFile();
					if (file != null && file.exists()) {
						try {
//							Integer lineno = rData.getPosition().line();
//							if (! l.contains(lineno)) {
//								l.add(lineno);
								Map map = new HashMap();
								MarkerUtilities.setLineNumber(map, rData.getPosition().line());
								MarkerUtilities.setMessage(map, "impact");
								MarkerUtilities.createMarker(file, map, ImpactConstants.IMPACT_MARKER);
//							}
						} catch (CoreException e) {
							e.printStackTrace();
						}
					}
				}
			}
		};
		visitor.breadthTraverse();
	}

	public static void deleteAllMarkers(final IProject project) {	
		try {
			project.deleteMarkers(ImpactConstants.IMPACT_MARKER,
					true, IResource.DEPTH_INFINITE);
		} catch (CoreException cEx) {}
	}
	
	private static void createTreeData(IProject project) {
	
		Map<String/*package name*/, Map<Aspect, AspectImpact>> packageImpactMap = abc.impact.ImpactDatabase.getPackageImpactMap();

		ITreeNode root;

		invisibleRoot = new InvisibleRootTreeNode();
		root = invisibleRoot;
		if (project != null) {
			IJavaProject jProject = JavaCore.create(project);
			if (jProject != null) {
				ITreeNode projectNode = new ProjectTreeNode(invisibleRoot, jProject);
				root = projectNode;
			}
		}
		
		for (Entry<String, Map<Aspect, AspectImpact>> pkgEntry : packageImpactMap.entrySet()) {
			ITreeNode packageNode = new PackageTreeNode(root, pkgEntry.getKey());
			
			for (Entry<Aspect, AspectImpact> apEntry : pkgEntry.getValue().entrySet()) {
				ITreeNode aspectNode = new AspectTreeNode(packageNode, apEntry.getKey());

				for (Entry<InAspect, InAspectImpact> iiEntry : apEntry.getValue().getInAspectImpactMap().entrySet()) {
					ITreeNode inAspectNode = new InAspectTreeNode(aspectNode, iiEntry.getKey());
					
					InAspectImpact iiImpact = iiEntry.getValue();

					Set<ShadowingImpact> shadowSet = iiImpact.getShadowingImpact();
					if (! shadowSet.isEmpty()) {
						ITreeNode shadowingType = new ShadowingStringTreeNode(inAspectNode);
						for (ShadowingImpact shadowingImpact : shadowSet) {
							for (SootClass affectedType : shadowingImpact.getAffectedTypes()) {
								ITreeNode fNode = new ShadowingImpactAffectedTypeTreeNode(shadowingType, shadowingImpact, affectedType);
								new ShadowingImpactOriginTypeTreeNode(fNode, shadowingImpact, shadowingImpact.getOriginType());
								new ShadowingImpactCurrentTypeTreeNode(fNode, shadowingImpact, shadowingImpact.getCurrentType());
							}
						}
					}
					
					Set<LookupImpact> lookSet = iiImpact.getLookupImpact();
					if (! lookSet.isEmpty()) {
						ITreeNode lookupType = new LookupStringTreeNode(inAspectNode);
						for (LookupImpact lookupImpact : lookSet) {
							ITreeNode mNode = new LookupImpactAffectedTypeTreeNode(lookupType, lookupImpact);
							for (LookupChange lookupChange : lookupImpact.getLookupChanges()) {
								ITreeNode placeNode = new ReferredPlaceTreeNode(mNode, lookupChange.getInvocationPlace());
								new LookupImpactOriginTypeTreeNode(placeNode, lookupImpact, lookupChange);
								new LookupImpactCurrentTypeTreeNode(placeNode, lookupImpact, lookupChange);
							}
						}
					}
				} // inaspect end
				
				for (Entry<AdviceDecl, AdviceImpact> adEntry : apEntry.getValue().getAdviceImpactMap().entrySet()) {
					ITreeNode adviceNode = new AdviceDeclTreeNode(aspectNode, adEntry.getKey());
					
					AdviceImpact adImpact = adEntry.getValue();
					
					ITreeNode compType = new ComputationStringTreeNode(adviceNode);  
					new ComputationImpactTreeNode(compType, adImpact.getComputationImpact());
					
					Set<StateImpact> stateSet = adImpact.getStateImpact();
					if (! stateSet.isEmpty()) {
						ITreeNode stateType = new StateStringTreeNode(adviceNode);
						for (StateImpact stateImpact : stateSet) {
							ITreeNode adviceStmtNode = new StateImpactAdviceStmtTreeNode(stateType, stateImpact);
							
							Set<Mutation> d = stateImpact.getDirectMutations();
							if (!d.isEmpty()) {
								ITreeNode directNode = new DirectStateStringTreeNode(adviceStmtNode);
								for (Mutation mutation : d) {
									createMutationSubTree(directNode, mutation);
								}
							}
							
							Set<Mutation> ind = stateImpact.getIndirectMutations();
							if (!ind.isEmpty()) {
								ITreeNode inDirectNode = new IndirectStateStringTreeNode(adviceStmtNode);
								for (Mutation mutation : ind) {
									ITreeNode muStmt = new MutationMutateStmtTreeNode(inDirectNode, mutation);
									createMutationSubTree(muStmt, mutation);
								}
							}
						}
					}
				} // advice end
			}
		}
	}

	private static void createMutationSubTree(ITreeNode parentNode, Mutation mutation) {
		for (Type type : mutation.getMutatedTypes()) {
			ITreeNode muField = new MutationMutatedTypeFieldTreeNode(parentNode, mutation, type);
			if (mutation.getMutatedFieldDeclClass() != null) {
				new MutationMutatedTypeFieldDeclClassTreeNode(muField, mutation);
			}
		}
	}

	public static List<ITreeNode> getSelectedNodes() {
		return selectedNodes;
	}
}
